"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LOGOUT_MUTATION = exports.REFRESH_TOKEN_MUTATION = void 0;
const core_1 = require("@apollo/client/core");
const utils_1 = require("@apollo/client/link/utils");
const utilities_1 = require("@apollo/client/utilities");
const MemoryStorage_1 = __importDefault(require("./storage/MemoryStorage"));
const REFRESH_TOKEN_KEY = ':auth:refreshToken';
const REFRESH_TOKEN_EXPIRES_KEY = ':auth:refreshTokenExpires';
const ACCESS_TOKEN_KEY = ':auth:accessToken';
const ACCESS_TOKEN_EXPIRES_KEY = ':auth:accessTokenExpires';
const DEFAULT_NAMESPACE = 'slicknode';
exports.REFRESH_TOKEN_MUTATION = core_1.gql `mutation refreshToken($token: String!) {
  refreshAuthToken(input: {refreshToken: $token}) {
    accessToken
    refreshToken
    accessTokenLifetime
    refreshTokenLifetime
  }
}`;
exports.LOGOUT_MUTATION = core_1.gql `mutation logout($refreshToken: String) {
  logoutUser(input:{refreshToken:$refreshToken}) {
  	success
	}
}`;
const authenticationDirectiveRemoveConfig = {
    test: (directive) => directive.name.value === 'authenticate',
    remove: false,
};
/**
 * SlicknodeLink instance to be used to load data with apollo-client
 * from slicknode GraphQL servers
 */
class SlicknodeLink extends core_1.ApolloLink {
    /**
     * Constructor
     * @param options
     */
    constructor(options = {}) {
        super();
        this.options = options;
        this.namespace = options.namespace || DEFAULT_NAMESPACE;
        this.storage = options.storage || global.localStorage || new MemoryStorage_1.default();
    }
    /**
     *
     * @param {Operation} operation
     * @param {NextLink} forward
     * @returns {Observable<FetchResult> | null}
     */
    request(operation, forward) {
        if (!forward) {
            throw new Error('Network link is missing in apollo client or SlicknodeLink is last link in the chain.');
        }
        return new core_1.Observable((observer) => {
            this.getAuthHeaders(forward)
                .then((authHeaders) => {
                operation.setContext(({ headers }) => ({
                    headers: Object.assign(Object.assign({}, (headers || {})), authHeaders),
                }));
                const definitions = operation.query.definitions;
                // Find current operation in definitions
                const currentOperation = definitions.find((operationDefinition) => {
                    return (operationDefinition.kind === 'OperationDefinition' &&
                        ((operationDefinition.name && operationDefinition.name.value === operation.operationName) ||
                            !operationDefinition.name));
                });
                // Check mutations for directives and logoutMutation
                const resultListeners = [];
                if (currentOperation && currentOperation.operation === 'mutation') {
                    const fields = [];
                    currentOperation.selectionSet.selections.forEach((selectionNode) => {
                        if (selectionNode.kind === 'Field') {
                            fields.push(selectionNode);
                        }
                        else {
                            // @TODO: Collect all relevant fields recursively from fragments / child fragments
                            // tslint:disable-next-line no-console
                            console.warn('Fragments not supported on Mutation type for slicknode-apollo-link');
                        }
                    });
                    fields.forEach((field) => {
                        if (field.name.value === 'logoutUser') {
                            // Subscribe to result to remove auth tokens from storage
                            resultListeners.push(() => {
                                this.debug('Removing auth tokens from storage');
                                this.logout();
                            });
                        }
                        else if (field.directives &&
                            field.directives.find((directive) => directive.name.value === 'authenticate')) {
                            const fieldName = field.alias ? field.alias.value : field.name.value;
                            // Subscribe to result to set auth token set
                            resultListeners.push((result) => {
                                // Validate auth token set and update tokens if valid
                                if (result.data &&
                                    result.data.hasOwnProperty(fieldName) &&
                                    typeof result.data[fieldName] === 'object') {
                                    const tokenSet = result.data[fieldName];
                                    this.validateAndSetAuthTokenSet(tokenSet);
                                }
                                else {
                                    this.debug('No valid token set returned');
                                }
                            });
                        }
                    });
                }
                // Remove @authenticated directives from document
                operation.query = utilities_1.removeDirectivesFromDocument([authenticationDirectiveRemoveConfig], operation.query);
                const nextObservable = forward(operation);
                // Add result listeners for token and logout processing
                nextObservable.subscribe({
                    complete() {
                        observer.complete();
                    },
                    error(errorValue) {
                        observer.error(errorValue);
                    },
                    next(value) {
                        resultListeners.forEach((listener) => listener(value));
                        observer.next(value);
                    },
                });
            })
                .catch((error) => {
                this.debug('Error obtaining auth headers in SlicknodeLink');
                observer.error(error);
            });
        });
    }
    /**
     * Returns true if the client has a valid access token
     *
     * @returns {boolean}
     */
    hasAccessToken() {
        return Boolean(this.getAccessToken());
    }
    /**
     * Returns true if the client has a valid refresh token
     *
     * @returns {boolean}
     */
    hasRefreshToken() {
        return Boolean(this.getRefreshToken());
    }
    /**
     * Updates the auth token set
     * @param token
     */
    setAuthTokenSet(token) {
        this.setAccessToken(token.accessToken);
        this.setAccessTokenExpires(token.accessTokenLifetime * 1000 + Date.now());
        this.setRefreshToken(token.refreshToken);
        this.setRefreshTokenExpires(token.refreshTokenLifetime * 1000 + Date.now());
    }
    /**
     * Stores the refreshToken in the storage of the client
     * @param token
     */
    setRefreshToken(token) {
        const key = this.namespace + REFRESH_TOKEN_KEY;
        this.storage.setItem(key, token);
    }
    /**
     * Returns the refresh token, NULL if none was stored yet
     * @returns {string|null}
     */
    getRefreshToken() {
        if ((this.getRefreshTokenExpires() || 0) < Date.now()) {
            return null;
        }
        const key = this.namespace + REFRESH_TOKEN_KEY;
        return this.storage.getItem(key);
    }
    /**
     * Sets the time when the auth token expires
     */
    setAccessTokenExpires(timestamp) {
        const key = this.namespace + ACCESS_TOKEN_EXPIRES_KEY;
        if (timestamp) {
            this.storage.setItem(key, String(timestamp));
        }
        else {
            this.storage.removeItem(key);
        }
    }
    /**
     * Returns the UNIX Timestamp when the refresh token expires
     * @returns {number|null}
     */
    getRefreshTokenExpires() {
        const key = this.namespace + REFRESH_TOKEN_EXPIRES_KEY;
        const expires = this.storage.getItem(key);
        return expires ? parseInt(expires, 10) : null;
    }
    /**
     * Sets the time when the auth token expires
     */
    setRefreshTokenExpires(timestamp) {
        const key = this.namespace + REFRESH_TOKEN_EXPIRES_KEY;
        this.storage.setItem(key, String(timestamp));
    }
    /**
     * Returns the UNIX Timestamp when the access token expires
     * @returns {number|null}
     */
    getAccessTokenExpires() {
        const key = this.namespace + ACCESS_TOKEN_EXPIRES_KEY;
        const expires = this.storage.getItem(key) || null;
        return expires ? parseInt(expires, 10) : null;
    }
    /**
     * Writes the access token to storage
     * @param token
     */
    setAccessToken(token) {
        const key = this.namespace + ACCESS_TOKEN_KEY;
        this.storage.setItem(key, token);
    }
    /**
     * Returns the access token, NULL if no valid token was found
     * @returns {null}
     */
    getAccessToken() {
        // Check if is expired
        if ((this.getAccessTokenExpires() || 0) < Date.now()) {
            return null;
        }
        const key = this.namespace + ACCESS_TOKEN_KEY;
        return this.storage.getItem(key) || null;
    }
    /**
     * Clears all tokens in the storage
     */
    logout() {
        return __awaiter(this, void 0, void 0, function* () {
            this.storage.removeItem(this.namespace + REFRESH_TOKEN_KEY);
            this.storage.removeItem(this.namespace + REFRESH_TOKEN_EXPIRES_KEY);
            this.storage.removeItem(this.namespace + ACCESS_TOKEN_KEY);
            this.storage.removeItem(this.namespace + ACCESS_TOKEN_EXPIRES_KEY);
        });
    }
    /**
     * Returns the headers that are required to authenticate at the GraphQL endpoint.
     * If no access tokens are available, an attempt is made to retrieve it from the backend
     * with the refreshToken
     */
    getAuthHeaders(forward) {
        // Check if headers are already being loaded
        if (!this.loadHeadersPromise) {
            const accessToken = this.options.accessToken || this.getAccessToken();
            if (accessToken) {
                this.debug('Using valid access token');
                return Promise.resolve({
                    Authorization: `Bearer ${accessToken}`,
                });
            }
            this.loadHeadersPromise = new Promise((resolve, reject) => {
                const refreshToken = this.getRefreshToken();
                // We have no token, try to get it from API via next link
                if (!accessToken && refreshToken) {
                    this.debug('No valid access token found, obtaining new AuthTokenSet with refresh token');
                    const refreshOperation = utils_1.createOperation({}, {
                        query: exports.REFRESH_TOKEN_MUTATION,
                        variables: {
                            token: refreshToken,
                        },
                    });
                    const observer = forward(refreshOperation);
                    observer.subscribe({
                        error: (error) => {
                            this.debug(`Error refreshing AuthTokenSet: ${error.message}`);
                            this.logout();
                            this.loadHeadersPromise = null;
                            resolve({});
                        },
                        next: (result) => {
                            this.loadHeadersPromise = null;
                            if (result.data && result.data.refreshAuthToken) {
                                if (!this.validateAndSetAuthTokenSet(result.data.refreshAuthToken)) {
                                    this.logout();
                                }
                            }
                            else {
                                this.debug('Refreshing auth token mutation failed');
                                this.logout();
                            }
                            resolve(this.hasAccessToken() ? {
                                Authorization: `Bearer ${this.getAccessToken()}`,
                            } : {});
                        },
                    });
                }
                else {
                    // Run in next event loop so loadHeadersPromise is reset after being assigned
                    setTimeout(() => {
                        this.loadHeadersPromise = null;
                        resolve({});
                    }, 0);
                }
            });
        }
        else {
            this.debug('Wait for token refresh in other transaction');
        }
        return this.loadHeadersPromise;
    }
    validateAndSetAuthTokenSet(tokenSet) {
        if (tokenSet &&
            typeof tokenSet === 'object' &&
            typeof tokenSet.accessToken === 'string' &&
            typeof tokenSet.accessTokenLifetime === 'number' &&
            typeof tokenSet.refreshToken === 'string' &&
            typeof tokenSet.refreshTokenLifetime === 'number') {
            // Update auth tokens in storage of link
            this.setAuthTokenSet(tokenSet);
            this.debug('Login successful, auth token set updated');
            return true;
        }
        this.debug('The auth token set has no valid format');
        return false;
    }
    debug(message) {
        if (this.options.debug) {
            console.log(`[Slicknode Auth] ${message}`); // tslint:disable-line no-console
        }
    }
}
exports.default = SlicknodeLink;
//# sourceMappingURL=SlicknodeLink.js.map